package state51::Odoo;
use Moose;

use Frontier::Client;
use WWW::Mechanize;
use state51::Logger;

our $VERSION = '2022102800';

=head1 NAME
state51::Odoo - library for working with Odoo

=head1 DESCRIPTION
 state51::Odoo is a library for working with Odoo

=head1 SYNOPSIS
    use state51::Odoo;

    my $odoo = state51::Odoo->new();
    my $uid = $odoo->odoo_uid;

    my $got = $odoo->execute(
            'account.move',
            'search',
            [
                [ 'ref', '>=',   $self->invoice_prefix . $invoice_number ],
                [ 'ref', 'like', $prefix . "%" ]
            ]
    );

=cut

use constant ODOO_TIMEOUT_SEC => $ENV{ODOO_TIMEOUT_SEC} // 180;

has user => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        return $ENV{ODOO_USERNAME} || die('No ODOO_USERNAME environment');;
    },
);

has password => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        return $ENV{ODOO_PASSWORD} || die('No ODOO_PASSWORD environment');
    },
);

has db_name => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        return $ENV{ODOO_DB} || die('No ODOO_DB environment');
    },
);

has base_url => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        my $base = $ENV{ODOO_URL} || die('No ODOO_URL environment');
        $base =~ s/\/$//;
        return $base;
    },
);

has logger => (
    is  => 'ro',
    isa => 'state51::Logger',
    default => sub {
        state51::Logger->new();
    },
);

has common_service => (
    is      => 'ro',
    isa     => 'Frontier::Client',
    lazy    => 1,
    builder => 'build_common_service',
);

=head2 build_common_service

Lazy builder to instantiate xml_rpc client instance for common operations.

=cut

sub build_common_service {
    my ($self) = @_;

    return $self->build_service('/xmlrpc/common');
}

has object_service => (
    is      => 'ro',
    isa     => 'Frontier::Client',
    lazy    => 1,
    builder => 'build_object_service',
);

=head2 build_object_service

Lazy builder to instantiate xml_rpc client instance for object operations.

=cut

sub build_object_service {
    my ($self) = @_;

    return $self->build_service('/xmlrpc/object');
}

sub build_service {
    my ($self, $path) = @_;

    my $server_url = $self->base_url . $path;

    my $cl = Frontier::Client->new('url' => $server_url);
    $cl->{ua}{timeout} = ODOO_TIMEOUT_SEC;

    return $cl;
}

has odoo_uid => (
    is            => 'ro',
    isa           => 'Int',
    lazy          => 1,
    builder       => 'login',
    documentation => 'logged in user_id',
    clearer       => 'odoo_sign_out',
);

=head2 login

Logging in to ODOO system. Would return an uid on success and a false
Frontier::RCP2::Boolean on failure.

=cut

sub login {
    my ($self) = @_;

    my $res = $self->common_service->call(
        'login',
        $self->db_name,
        $self->user,
        $self->password
    );

    if (ref $res eq 'Frontier::RPC2::Boolean') {
        $self->logger->message(
            level => 'warn',
            message => [
                '[Odoo::login] '
                . 'Something is wrong with the Odoo login, check the parameters.'
            ]
        ) if not $res->value;
    }

    return $res;
}

has report_service => (
    is      => 'ro',
    isa     => 'Frontier::Client',
    lazy    => 1,
    builder => 'build_report_service',
);

=head2 build_report_service

Lazy builder to instantiate xml_rpc client instance for report operations.

=cut

sub build_report_service {
    my ($self) = @_;

    my $server_url = $self->base_url . '/report';
    return Frontier::Client->new('url' => $server_url);
}

has www_mechanize => (
    is      => 'rw',
    isa     => 'WWW::Mechanize',
    lazy    => 1,
    builder => 'build_www_mechanize',
);

sub build_www_mechanize {
    my ($self) = @_;

    my $mech = WWW::Mechanize->new(
        onerror => undef,
        agent   => 'Linux Mozilla'
    );

    my $login_url = $self->base_url . "/web/login";
    $mech->get($login_url);

    if ($mech->res->is_success) {
        my $csrf_token = $mech->value('csrf_token');

        my $resp = $mech->submit_form(
            with_fields => {
                "login"      => $self->user,
                "password"   => $self->password,
                "csrf_token" => $csrf_token,
            }
        );

        unless ($resp->is_success) {
            die "Web login failed";
        }
    }
    else {
        die $mech->content;
    }

    return $mech;
}

sub download_invoice_report {
    my ($self, $report_filename, $odoo_invoice_id) = @_;

    my $report_path = '/report/pdf/account.report_invoice/' . $odoo_invoice_id; 
    my $report_url = $self->base_url . $report_path;

    my $response = $self->www_mechanize->get($report_url, ":content_file" => $report_filename);

    unless ($response->is_success) {
        $self->logger->message(
            level => 'warn',
            message => [
                '[Odoo::download_invoice_report] '
                . 'The invoice report file could not be downloaded. '
                . 'Code: ' . $response->code
                . 'Message: ' . $response->message
            ]
        );

        return;
    }

    return $response;
}

sub execute {
    my ($self, $model, $method, $position, $keywords, $args) = @_;

    unless ($model && $method) {
        $self->logger->message(
            level => 'warn',
            message => [
                '[Odoo::execute] Model or method is empty',
            ]
        );

        return;
    }

    my $result = eval {
        $self->object_service->call(
            'execute', $self->db_name, $self->odoo_uid, $self->password,
            $model, $method, $position,
            $keywords ? $keywords : (),
        );
    };

    if ($@) {
        $self->logger->message(
            level => 'fatal',
            message => ['[Odoo::execute] ' . $@],
        );

        return;
    }

    if ($method eq 'read' and $args->{replace_boolean}) {
        for my $item (@$result) {
            for my $k (keys %$item) {
                if (ref $item->{$k} eq 'Frontier::RPC2::Boolean') {
                    $item->{$k} = $item->{$k}->value || undef;
                }
            }
        }
    }

    return $result;
}

1;
